# from __future__ import absolute_import

from spooq import extractor
from spooq import loader
from spooq import pipeline
from spooq import transformer
from spooq.pipeline import PipelineFactory
from . import spooq_logger

if not spooq_logger.initialized:
    spooq_logger.initialize()

__all__ = ["extractor", "transformer", "loader", "pipeline", "PipelineFactory"]



"""
To decrease the complexity of building data pipelines for data engineers, an expert system or
business rules engine can be used to automatically build and configure a data pipeline based on
context variables, groomed metadata, and relevant rules.
"""
from __future__ import print_function

from builtins import object
import requests
import json
from spooq.pipeline import Pipeline
import spooq.extractor as E
import spooq.transformer as T
import spooq.loader as L


class PipelineFactory(object):
    """
    Provides an interface to automatically construct pipelines for Spooq.

    Example
    -------
    >>> pipeline_factory = PipelineFactory()
    >>>
    >>> #  Fetch user data set with applied mapping, filtering,
    >>> #  and cleaning transformers
    >>> df = pipeline_factory.execute({
    >>>      "entity_type": "user",
    >>>      "date": "2018-10-20",
    >>>      "time_range": "last_day"})
    >>>
    >>> #  Load user data partition with applied mapping, filtering,
    >>> #  and cleaning transformers to a hive database
    >>> pipeline_factory.execute({
    >>>      "entity_type": "user",
    >>>      "date": "2018-10-20",
    >>>      "batch_size": "daily"})

    Attributes
    ----------
    url : :any:`str`, (Defaults to "http://localhost:5000/pipeline/get")
        The end point of an expert system which will be called to infer names and parameters.

    Note
    ----
    PipelineFactory is only responsible for querying an expert system with provided parameters
    and constructing a Spooq pipeline out of the response. It does not have any reasoning capabilities
    itself! It requires therefore a HTTP service responding with a JSON object containing following structure:

    ::

        {
            "extractor": {"name": "Type1Extractor", "params": {"key 1": "val 1", "key N": "val N"}},
            "transformers": [
                {"name": "Type1Transformer", "params": {"key 1": "val 1", "key N": "val N"}},
                {"name": "Type2Transformer", "params": {"key 1": "val 1", "key N": "val N"}},
                {"name": "Type3Transformer", "params": {"key 1": "val 1", "key N": "val N"}},
                {"name": "Type4Transformer", "params": {"key 1": "val 1", "key N": "val N"}},
                {"name": "Type5Transformer", "params": {"key 1": "val 1", "key N": "val N"}},
            ],
            "loader": {"name": "Type1Loader", "params": {"key 1": "val 1", "key N": "val N"}}
        }

    Hint
    ----
    There is an experimental implementation of an expert system which complies with the requirements
    of PipelineFactory called `spooq_rules`. If you are interested, please ask the author of Spooq about it.
    """

    def __init__(self, url="http://localhost:5000/pipeline/get"):
        self.url = url

    def execute(self, context_variables):
        """
        Fetches a ready-to-go pipeline instance via :py:meth:`get_pipeline`
        and executes it.

        Parameters
        ----------
            context_variables : :py:class:`dict`
                These collection of parameters should describe the current context about the use case
                of the pipeline. Please see the examples of the PipelineFactory class'
                documentation.

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            If the loader component is by-passed (in the case of ad_hoc use cases).

        :any:`None`
            If the loader component does not return a value (in the case of persisting data).
        """
        pipeline = self.get_pipeline(context_variables)
        return pipeline.execute()

    def get_metadata(self, context_variables):
        """
        Sends a POST request to the defined endpoint (`url`) containing the
        supplied context variables.

        Parameters
        ----------
        context_variables : :py:class:`dict`
            These collection of parameters should describe the current context about the use case
            of the pipeline. Please see the examples of the PipelineFactory class'
            documentation.

        Returns
        -------
        :py:class:`dict`
            Names and parameters of each ETL component to construct a Spooq pipeline
        """
        return requests.post(self.url, json=context_variables).json()

    def get_pipeline(self, context_variables):
        """
        Fetches the necessary metadata via :py:meth:`get_metadata` and
        returns a ready-to-go pipeline instance.

        Parameters
        ----------
        context_variables : :py:class:`dict`
            These collection of parameters should describe the current context about the use case
            of the pipeline. Please see the examples of the PipelineFactory class'
            documentation.

        Returns
        -------
        :py:class:`~spooq.Pipeline`
            A Spooq pipeline instance which is fully configured and can still be
            adapted and consequently executed.
        """
        metadata = self.get_metadata(context_variables)
        pipeline = Pipeline()
        extractor = self._get_extractor(metadata)
        pipeline.set_extractor(extractor)
        transformers = self._get_transformers(metadata)
        pipeline.add_transformers(transformers)
        loader = self._get_loader(metadata)
        if not loader:
            pipeline.bypass_loader = True
        else:
            pipeline.set_loader(loader)
        return pipeline

    def _get_extractor(self, magic_data):
        extractor_class = getattr(E, magic_data.get("extractor", {}).get("name", ""))
        extractor_params = magic_data.get("extractor", {}).get("params", "")
        return extractor_class(**extractor_params)

    def _get_transformers(self, magic_data):
        transformers = []
        for transformer in magic_data["transformers"]:
            transformer_class = getattr(T, transformer["name"])
            print(transformer_class)
            transformers.append(transformer_class(**transformer["params"]))
        return transformers

    def _get_loader(self, magic_data):
        loader_name = magic_data.get("loader", {}).get("name", "")
        if loader_name == "ByPass":
            return False
        loader_class = getattr(L, loader_name)
        loader_params = magic_data.get("loader", {}).get("params", "")
        return loader_class(**loader_params)



from __future__ import absolute_import
from builtins import str
from past.builtins import basestring
import pandas as pd
from copy import copy
from pyspark.sql import SparkSession
from pyspark.sql.functions import min as sql_min
from pyspark.sql.functions import max as sql_max
from pyspark.sql.types import StructField, StructType
from pyspark.sql.types import IntegerType, StringType


from .extractor import Extractor


class JDBCExtractor(Extractor):
    def __init__(self, jdbc_options, cache=True):
        super(JDBCExtractor, self).__init__()
        self._assert_jdbc_options(jdbc_options)
        self.jdbc_options = jdbc_options
        self.cache = cache
        self.spark = (
            SparkSession.Builder()
            .enableHiveSupport()
            .appName("spooq.extractor: {nm}".format(nm=self.name))
            .getOrCreate()
        )

    def _load_from_jdbc(self, query, jdbc_options, cache=True):
        jdbc_options = copy(self.jdbc_options)
        jdbc_options["dbtable"] = "({q}) as table_statement".format(q=query)
        source_df = self.spark.read.format("jdbc").options(**jdbc_options).load()
        if cache:
            source_df.cache()

        return source_df

    def _assert_jdbc_options(self, jdbc_options):
        for key in ["url", "driver", "user", "password"]:
            assert key in jdbc_options, key + " is missing from the jdbc_options."
            assert isinstance(jdbc_options[key], basestring), key + " has to be provided as a string object."


class JDBCExtractorFullLoad(JDBCExtractor):
    """
    Connects to a JDBC Source and fetches the data defined by the provided Query.

    Examples
    --------
    >>> import spooq.extractor as E
    >>>
    >>> extractor = E.JDBCExtractorFullLoad(
    >>>     query="select id, first_name, last_name, gender, created_at test_db.from users",
    >>>     jdbc_options={
    >>>         "url": "jdbc:postgresql://localhost/test_db",
    >>>         "driver": "org.postgresql.Driver",
    >>>         "user": "read_only",
    >>>         "password": "test123",
    >>>     },
    >>> )
    >>>
    >>> extracted_df = extractor.extract()
    >>> type(extracted_df)
    pyspark.sql.dataframe.DataFrame

    Parameters
    ----------
    query : :any:`str`
        Defines the actual query sent to the JDBC Source. This has to be a valid SQL query
        with respect to the source system (e.g., T-SQL for Microsoft SQL Server).

    jdbc_options : :class:`dict`, optional
        A set of parameters to configure the connection to the source:
            * **url** (:any:`str`) - A JDBC URL of the form jdbc:subprotocol:subname.
              e.g., jdbc:postgresql://localhost:5432/dbname
            * **driver** (:any:`str`) - The class name of the JDBC driver to use to connect to this URL.
            * **user** (:any:`str`) - Username to authenticate with the source database.
            * **password** (:any:`str`) - Password to authenticate with the source database.

        See :meth:`pyspark.sql.DataFrameReader.jdbc` and
        https://spark.apache.org/docs/2.4.3/sql-data-sources-jdbc.html for more information.

    cache : :any:`bool`, defaults to :any:`True`
        Defines, weather to :meth:`~pyspark.sql.DataFrame.cache` the dataframe, after it is loaded.
        Otherwise the Extractor will reload all data from the source system eachtime an action is
        performed on the DataFrame.

    Raises
    ------
    :any:`exceptions.AssertionError`:
        All jdbc_options values need to be present as string variables.
    """

    def __init__(self, query, jdbc_options, cache=True):
        super(JDBCExtractorFullLoad, self).__init__(jdbc_options=jdbc_options, cache=cache)
        self.query = query

    def extract(self):
        """
        This is the Public API Method to be called for all classes of Extractors

        Parameters
        ----------

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            PySpark dataframe from the input JDBC connection.

        """
        return self._load_from_jdbc(self.query, jdbc_options=self.jdbc_options, cache=self.cache)


class JDBCExtractorIncremental(JDBCExtractor):
    """
    Connects to a JDBC Source and fetches the data with respect to boundaries.
    The boundaries are inferred from the partition to load and logs from previous loads
    stored in the ``spooq_values_table``.

    Examples
    --------
    >>> import spooq.extractor as E
    >>>
    >>> # Boundaries derived from previously logged extractions => ("2020-01-31 03:29:59", False)
    >>>
    >>> extractor = E.JDBCExtractorIncremental(
    >>>     partition="20200201",
    >>>     jdbc_options={
    >>>         "url": "jdbc:postgresql://localhost/test_db",
    >>>         "driver": "org.postgresql.Driver",
    >>>         "user": "read_only",
    >>>         "password": "test123",
    >>>     },
    >>>     source_table="users",
    >>>     spooq_values_table="spooq_jdbc_log_users",
    >>> )
    >>>
    >>> extractor._construct_query_for_partition(extractor.partition)
    select * from users where updated_at > "2020-01-31 03:29:59"
    >>>
    >>> extracted_df = extractor.extract()
    >>> type(extracted_df)
    pyspark.sql.dataframe.DataFrame

    Parameters
    ----------
    partition : :any:`int` or :any:`str`
        Partition to extract. Needed for logging the incremental load in
        the ``spooq_values_table``.

    jdbc_options : :class:`dict`, optional
        A set of parameters to configure the connection to the source:
            * **url** (:any:`str`) - A JDBC URL of the form jdbc:subprotocol:subname.
              e.g., jdbc:postgresql://localhost:5432/dbname
            * **driver** (:any:`str`) - The class name of the JDBC driver to use to connect to this URL.
            * **user** (:any:`str`) - Username to authenticate with the source database.
            * **password** (:any:`str`) - Password to authenticate with the source database.

        See :meth:`pyspark.sql.DataFrameReader.jdbc` and
        https://spark.apache.org/docs/2.4.3/sql-data-sources-jdbc.html for more information.

    source_table : :any:`str`
        Defines the tablename of the source to be loaded from. For example 'purchases'.
        This is necessary to build the query.

    spooq_values_table : :any:`str`
        Defines the Hive table where previous and future loads of a specific source table
        are logged. This is necessary to derive boundaries for the current partition.

    spooq_values_db : :any:`str`, optional
        Defines the Database where the ``spooq_values_table`` is stored.
        Defaults to `'spooq_values'`.

    spooq_values_partition_column : :any:`str`, optional
        The column name which is used for the boundaries.
        Defaults to `'updated_at'`.

    cache : :any:`bool`, defaults to :any:`True`
        Defines, weather to :meth:`~pyspark.sql.DataFrame.cache` the dataframe, after it is loaded. Otherwise the Extractor
        will reload all data from the source system again, if a second action upon the dataframe
        is performed.

    Raises
    ------
    :any:`exceptions.AssertionError`:
        All jdbc_options values need to be present as string variables.
    """

    def __init__(
        self,
        partition,
        jdbc_options,
        source_table,
        spooq_values_table,
        spooq_values_db="spooq_values",
        spooq_values_partition_column="updated_at",
        cache=True,
    ):
        super(JDBCExtractorIncremental, self).__init__(jdbc_options)
        self.partition = partition
        self.source_table = source_table
        self.spooq_values_table = spooq_values_table
        self.spooq_values_db = spooq_values_db
        self.spooq_values_partition_column = spooq_values_partition_column
        self.cache = cache

    def extract(self):
        query = self._construct_query_for_partition(partition=self.partition)
        loaded_df = self._load_from_jdbc(query, self.jdbc_options, cache=self.cache)
        self._update_boundaries_for_current_partition_on_table(
            loaded_df,
            self.spooq_values_db,
            self.spooq_values_table,
            self.partition,
            self.spooq_values_partition_column,
        )
        return loaded_df

    def _construct_query_for_partition(self, partition):
        """Constructs and returns a predicated Query :any:`str` depending on the `partition`

        Based on the partition and previous loading logs (`spooq_values_table`),
        boundaries will be calculated and injected in the where clause of the query.

        Parameters
        ----------
        partition : Integer or :any:`str`

        Returns
        -------
        :any:`str`
            Complete Query :any:`str` to be used for JDBC Connections

        """

        select_statement = "select *"
        where_clause = ""
        lower_bound, upper_bound = self._get_boundaries_for_import(partition)

        def _fix_boundary_value_syntax(boundary):
            """If a boundary value is not a number, it has to be quoted for correct syntax."""

            try:
                boundary = int(boundary)
            except ValueError:
                boundary = '"{bnd}"'.format(bnd=boundary)
            return boundary

        if lower_bound and upper_bound:
            where_clause = "where {chk_col} > {low_bnd} and {chk_col} <= {up_bnd}".format(
                chk_col=self.spooq_values_partition_column,
                low_bnd=_fix_boundary_value_syntax(lower_bound),
                up_bnd=_fix_boundary_value_syntax(upper_bound),
            )
        elif lower_bound:
            where_clause = "where {chk_col} > {low_bnd}".format(
                chk_col=self.spooq_values_partition_column,
                low_bnd=_fix_boundary_value_syntax(lower_bound),
            )
        elif upper_bound:
            where_clause = "where {chk_col} <= {up_bnd}".format(
                chk_col=self.spooq_values_partition_column,
                up_bnd=_fix_boundary_value_syntax(upper_bound),
            )

        query = "{select} from {tbl} {where}".format(select=select_statement, tbl=self.source_table, where=where_clause)
        return " ".join(query.split())

    def _get_boundaries_for_import(self, partition):
        """
        Returns the lower and upper boundaries to be used in the where clause of the query.
        This information is deducted from the ``partition`` parameter and previous loading logs
        (persisted in `spooq_values_table`).

        Parameters
        ----------
        partition : :py:class:`int` or :any:`str`

        Returns
        -------
        Tuple of :any:`str`
            Values of the tuple can also be `False`
        """
        pd_df = self._get_previous_boundaries_table_as_pd(self.spooq_values_db, self.spooq_values_table)
        partition = int(partition)
        table_is_empty = pd_df.empty
        partition_exists = not pd_df.loc[pd_df["dt"] == partition].empty
        succeeding_partition_exists = not pd_df.loc[pd_df["dt"] > partition].empty
        preceding_partition_exists = not pd_df.loc[pd_df["dt"] < partition].empty

        if table_is_empty:
            """First import ever, starting from zero"""
            return False, False

        elif partition_exists:
            """Partition to insert already exists (reload / backfill)"""
            return self._get_lower_and_upper_bounds_from_current_partition(pd_df, partition)

        else:
            """Partition to insert does not yet exist (new day to insert)"""
            if preceding_partition_exists and not succeeding_partition_exists:
                """No equal or newer partitions exist (newest partition will be imported). Default Case"""
                return (
                    self._get_upper_bound_from_preceding_partition(pd_df, partition),
                    False,
                )

            elif not preceding_partition_exists and succeeding_partition_exists:
                """No older or equal partitions exist"""
                return (
                    False,
                    self._get_lower_bound_from_succeeding_partition(pd_df, partition),
                )

            elif preceding_partition_exists and succeeding_partition_exists:
                """At least one older, no equal and at least one newer partitions exist"""
                return (
                    self._get_upper_bound_from_preceding_partition(pd_df, partition),
                    self._get_lower_bound_from_succeeding_partition(pd_df, partition),
                )

            else:
                raise Exception(
                    """
                    ERROR: Something weird happened...
                    There was a logical problem getting the correct boundaries
                    from the spooqvalue table!
                    Please debug me ;-)
                    """
                )

    def _get_previous_boundaries_table_as_pd(self, spooq_values_db, spooq_values_table):
        """
        Converts the previous_boundaries_table and returns a Pandas Dataframe.

        Parameters
        ----------
        spooq_values_db : :any:`str`
        spooq_values_table : :any:`str`

        Returns
        -------
        Pandas Dataframe
            Content of `spooq_values_table` from `spooq_values_db`
        """
        return self._get_previous_boundaries_table(spooq_values_db, spooq_values_table).toPandas()

    def _get_previous_boundaries_table(self, spooq_values_db, spooq_values_table):
        """
        Fetches and returns a DataFrame containing the logs of previous loading
        jobs (`spooq_values_table`) of this entity

        Parameters
        ----------
        spooq_values_db : :any:`str`
        spooq_values_table : :any:`str`

        Returns
        -------
        PySpark DataFrame
            Content of `spooq_values_table` from `spooq_values_db`

        """
        table_name = "{db}.{tbl}".format(db=spooq_values_db, tbl=spooq_values_table)
        self.logger.info("Loading spooqValues Table from {name}".format(name=table_name))
        df = self.spark.table(table_name)
        try:
            self.spooq_values_partition_column
        except AttributeError:
            self.spooq_values_partition_column = df.select("partition_column").distinct().collect()[0].partition_column
        return df

    @staticmethod
    def _get_lower_bound_from_succeeding_partition(pd_df, partition):
        succeeding_pd_df = pd_df.loc[pd_df["dt"] > partition]
        return succeeding_pd_df.sort_values("dt", ascending=1).iloc[0].first_value

    @staticmethod
    def _get_upper_bound_from_preceding_partition(pd_df, partition):
        preceding_pd_df = pd_df.loc[pd_df["dt"] < partition]
        return preceding_pd_df.sort_values("dt", ascending=0).iloc[0].last_value

    @staticmethod
    def _get_lower_and_upper_bounds_from_current_partition(pd_df, partition):
        current_pd_df = pd_df.loc[pd_df["dt"] == partition].iloc[0]
        return (current_pd_df.first_value, current_pd_df.last_value)

    def _get_lowest_boundary_from_df(self, df, spooq_values_partition_column):
        return eval(
            "df.na.drop(how='any', subset=['{chk_col}']).select(sql_min(df.{chk_col}).alias('minimum')).collect()[0].minimum".format(
                chk_col=spooq_values_partition_column
            )
        )

    def _get_highest_boundary_from_df(self, df, spooq_values_partition_column):
        return eval(
            "df.na.drop(how='any', subset=['{chk_col}']).select(sql_max(df.{chk_col}).alias('maximum')).collect()[0].maximum".format(
                chk_col=spooq_values_partition_column
            )
        )

    def _update_boundaries_for_current_partition_on_table(
        self, df, spooq_values_db, spooq_values_table, partition, spooq_values_partition_column
    ):

        lowest_boundary = self._get_lowest_boundary_from_df(df, spooq_values_partition_column)
        highest_boundary = self._get_highest_boundary_from_df(df, spooq_values_partition_column)
        self._write_boundaries_to_hive(
            lowest_boundary,
            highest_boundary,
            spooq_values_db,
            spooq_values_table,
            partition,
            spooq_values_partition_column,
        )

    def _write_boundaries_to_hive(
        self,
        lowest_boundary,
        highest_boundary,
        spooq_values_db,
        spooq_values_table,
        partition,
        spooq_values_partition_column,
    ):

        self.spark.conf.set("hive.exec.dynamic.partition", "true")
        self.spark.conf.set("hive.exec.dynamic.partition.mode", "nonstrict")

        schema = StructType(
            [
                StructField("partition_column", StringType(), True),
                StructField("dt", IntegerType(), False),
                StructField("first_value", StringType(), False),
                StructField("last_value", StringType(), False),
            ]
        )

        input_data = [
            [
                str(spooq_values_partition_column),
                int(partition),
                str(lowest_boundary),
                str(highest_boundary),
            ]
        ]

        df_output = self.spark.createDataFrame(input_data, schema=schema)

        df_output.repartition(1).write.mode("overwrite").insertInto(
            "{db}.{tbl}".format(db=spooq_values_db, tbl=spooq_values_table, dt=partition)
        )



"""
Loaders take a :py:class:`pyspark.sql.DataFrame` as an input and save it to a sink.

Each Loader class has to have a `load` method which takes a DataFrame as single paremter.

Possible Loader sinks can be **Hive Tables**, **Kudu Tables**, **HBase Tables**, **JDBC
Sinks** or **ParquetFiles**.
"""

from builtins import object
import logging


class Loader(object):
    """
    Base Class of Loader Objects.

    Attributes
    ----------
    name : :any:`str`
        Sets the `__name__` of the class' type as `name`, which is essentially the Class' Name.
    logger : :any:`logging.Logger`
        Shared, class level logger for all instances.

    """

    def __init__(self):
        self.name = type(self).__name__
        self.logger = logging.getLogger("spooq")

    def load(self, input_df):
        """
        Persists data from a PySpark DataFrame to a target table.

        Parameters
        ----------
        input_df : :any:`pyspark.sql.DataFrame`
            Input DataFrame which has to be loaded to a target destination.

        Note
        ----
        This method takes only a single DataFrame as an input parameter. All other needed
        parameters are defined in the initialization of the Loader object.
        """
        raise NotImplementedError("This method has to be implemented in the subclasses")

    def __str__(self):
        return "Loader Object of Class {nm}".format(nm=self.name)



"""This is a collection of module level methods to construct a specific
PySpark DataFrame query for custom defined data types.

These methods are not meant to be called directly but via the
the :py:class:`~spooq.transformer.mapper.Mapper` transformer.
Please see that particular class on how to apply custom data types.

For injecting your **own custom data types**, please have a visit to the
:py:meth:`add_custom_data_type` method!
"""
from __future__ import division
import sys

if sys.version_info.major > 2:
    # This is needed for python 2 as otherwise pyspark raises an exception for following method:
    # _to_json
    # ToDo: Check if Python3 works without this import / overwrite
    from builtins import str

from past.utils import old_div
import sys
import json
from pyspark.sql import functions as F
from pyspark.sql import types as T

__all__ = ["add_custom_data_type"]

MIN_TIMESTAMP_MS = 0  # 01.01.1970 00:00:00
MAX_TIMESTAMP_MS = 4102358400000  # 31.12.2099 00:00:00

MIN_TIMESTAMP_SEC = 0
MAX_TIMESTAMP_SEC = old_div(MAX_TIMESTAMP_MS, 1000)


def add_custom_data_type(function_name, func):
    """
    Registers a custom data type in runtime to be used with the :py:class:`~spooq.transformer.mapper.Mapper` transformer.

    Example
    -------
    >>> import spooq.transformer.mapper_custom_data_types as custom_types
    >>> import spooq.transformer as T
    >>> from pyspark.sql import Row, functions as F, types as T

    >>> def hello_world(source_column, name):
    >>>     "A UDF (User Defined Function) in Python"
    >>>     def _to_hello_world(col):
    >>>         if not col:
    >>>             return None
    >>>         else:
    >>>             return "Hello World"
    >>>
    >>>     udf_hello_world = F.udf(_to_hello_world, T.StringType())
    >>>     return udf_hello_world(source_column).alias(name)
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(hello_from=u'tsivorn1@who.int'),
    >>>      Row(hello_from=u''),
    >>>      Row(hello_from=u'gisaksen4@skype.com')]
    >>> )
    >>>
    >>> custom_types.add_custom_data_type(function_name="hello_world", func=hello_world)
    >>> transformer = T.Mapper(mapping=[("hello_who", "hello_from", "hello_world")])
    >>> df = transformer.transform(input_df)
    >>> df.show()
    +-----------+
    |  hello_who|
    +-----------+
    |Hello World|
    |       null|
    |Hello World|
    +-----------+

    >>> def first_and_last_name(source_column, name):
    >>>     "A PySpark SQL expression referencing multiple columns"
    >>>     return F.concat_ws("_", source_column, F.col("attributes.last_name")).alias(name)
    >>>
    >>> custom_types.add_custom_data_type(function_name="full_name", func=first_and_last_name)
    >>>
    >>> transformer = T.Mapper(mapping=[
    >>>     ("first_name", "attributes.first_name", "StringType"),
    >>>     ("last_name",  "attributes.last_name",  "StringType"),
    >>>     ("full_name",  "attributes.first_name", "full_name"),
    >>> ])

    Parameters
    ----------
    function_name: :any:`str`
        The name of your custom data type

    func: compatible function
        The PySpark dataframe function which will be called on a column, defined in the mapping
        of the Mapper class.
        Required input parameters are ``source_column`` and ``name``.
        Please see the note about required input parameter of custom data types for more information!

    Note
    ----
    Required input parameter of custom data types:

    **source_column** (:py:class:`pyspark.sql.Column`) - This is where your logic will be applied.
      The mapper transformer takes care of calling this method with the right column so you can just
      handle it like an object which you would get from ``df["some_attribute"]``.

    **name** (:any:`str`) - The name how the resulting column will be named. Nested attributes are not
      supported. The Mapper transformer takes care of calling this method with the right column name.
    """

    if not function_name.startswith("_generate_select_expression_for_"):
        function_name = "_generate_select_expression_for_" + function_name
    current_module = sys.modules[__name__]
    setattr(current_module, function_name, func)


def _get_select_expression_for_custom_type(source_column, name, data_type):
    """ Internal method for calling functions dynamically """
    current_module = sys.modules[__name__]
    function_name = "_generate_select_expression_for_" + data_type
    try:
        function = getattr(current_module, function_name)
    except AttributeError as e:
        raise AttributeError(
            "Spooq could not find a Select Expression Generator Method \n"
            + "for the custom DataType: "
            + '"'
            + data_type
            + '"'
            + "\n"
            + "Original Error Message: "
            + str(e)
        )

    return function(source_column, name)


def _generate_select_expression_for_as_is(source_column, name):
    """ alias for `_generate_select_expression_without_casting` """
    return _generate_select_expression_without_casting(source_column, name)


def _generate_select_expression_for_keep(source_column, name):
    """ alias for `_generate_select_expression_without_casting` """
    return _generate_select_expression_without_casting(source_column, name)


def _generate_select_expression_for_no_change(source_column, name):
    """ alias for `_generate_select_expression_without_casting` """
    return _generate_select_expression_without_casting(source_column, name)


def _generate_select_expression_without_casting(source_column, name):
    """
    Returns a column without casting. This is especially useful if you need to
    keep a complex data type, like an array, list or a struct.

    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(friends=[Row(first_name=None, id=3993, last_name=None), Row(first_name=u'Ru\xf2', id=17484, last_name=u'Trank')]),
     Row(friends=[]),
     Row(friends=[Row(first_name=u'Daphn\xe9e', id=16707, last_name=u'Lyddiard'), Row(first_name=u'Ad\xe9la\xefde', id=17429, last_name=u'Wisdom')])]
    >>> mapping = [("my_friends", "friends", "as_is")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(my_friends=[Row(first_name=None, id=3993, last_name=None), Row(first_name=u'Ru\xf2', id=17484, last_name=u'Trank')]),
     Row(my_friends=[]),
     Row(my_friends=[Row(first_name=u'Daphn\xe9e', id=16707, last_name=u'Lyddiard'), Row(first_name=u'Ad\xe9la\xefde', id=17429, last_name=u'Wisdom')])]


    """
    return source_column.alias(name)


def _generate_select_expression_for_json_string(source_column, name):
    """
    Returns a column as json compatible string.
    Nested hierarchies are supported.
    The unicode representation of a column will be returned if an error occurs.

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(friends=[Row(first_name=None, id=3993, last_name=None), Row(first_name=u'Ru\xf2', id=17484, last_name=u'Trank')]),
     Row(friends=[]),
     Row(friends=[Row(first_name=u'Daphn\xe9e', id=16707, last_name=u'Lyddiard'), Row(first_name=u'Ad\xe9la\xefde', id=17429, last_name=u'Wisdom')])]    >>> mapping = [("friends_json", "friends", "json_string")]
    >>> mapping = [("friends_json", "friends", "json_string")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(friends_json=u'[{"first_name": null, "last_name": null, "id": 3993}, {"first_name": "Ru\\u00f2", "last_name": "Trank", "id": 17484}]'),
     Row(friends_json=None),
     Row(friends_json=u'[{"first_name": "Daphn\\u00e9e", "last_name": "Lyddiard", "id": 16707}, {"first_name": "Ad\\u00e9la\\u00efde", "last_name": "Wisdom", "id": 17429}]')]
    """

    def _to_json(col):
        if not col:
            return None
        try:
            if isinstance(col, list):
                return json.dumps([x.asDict(recursive=True) for x in col])
            else:
                return json.dumps(col.asDict(recursive=True))
        except (AttributeError, TypeError):
            return str(col)

    udf_to_json = F.udf(_to_json, T.StringType())
    return udf_to_json(source_column).alias(name)


def _generate_select_expression_for_timestamp_ms_to_ms(source_column, name):
    """
    This Constructor is used for unix timestamps. The values are cleaned
    next to casting and renaming.
    If the values are not between `01.01.1970` and `31.12.2099`,
    NULL will be returned.
    Cast to :any:`pyspark.sql.types.LongType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(time_sec=1581540839000),  # 02/12/2020 @ 8:53pm (UTC)
    >>>     Row(time_sec=-4887839000),    # Invalid!
    >>>     Row(time_sec=4737139200000)   # 02/12/2120 @ 12:00am (UTC)
    >>> ])
    >>>
    >>> mapping = [("unix_ts", "time_sec", "timestamp_ms_to_ms")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(unix_ts=1581540839000), Row(unix_ts=None), Row(unix_ts=None)]

    Note
    ----
    *input*  in **milli seconds**
    *output* in **milli seconds**
    """

    return (
        F.when(source_column.between(MIN_TIMESTAMP_MS, MAX_TIMESTAMP_MS), source_column)
        .otherwise(F.lit(None))
        .cast(T.LongType())
        .alias(name)
    )


def _generate_select_expression_for_timestamp_ms_to_s(source_column, name):
    """
    This Constructor is used for unix timestamps. The values are cleaned
    next to casting and renaming.
    If the values are not between `01.01.1970` and `31.12.2099`,
    NULL will be returned.
    Cast to :any:`pyspark.sql.types.LongType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(time_sec=1581540839000),  # 02/12/2020 @ 8:53pm (UTC)
    >>>     Row(time_sec=-4887839000),    # Invalid!
    >>>     Row(time_sec=4737139200000)   # 02/12/2120 @ 12:00am (UTC)
    >>> ])
    >>>
    >>> mapping = [("unix_ts", "time_sec", "timestamp_ms_to_s")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(unix_ts=1581540839), Row(unix_ts=None), Row(unix_ts=None)]

    Note
    ----
    *input*  in **milli seconds**
    *output* in **seconds**
    """

    return (
        F.when(
            source_column.between(MIN_TIMESTAMP_MS, MAX_TIMESTAMP_MS),
            source_column / 1000,
        )
        .otherwise(F.lit(None))
        .cast(T.LongType())
        .alias(name)
    )


def _generate_select_expression_for_timestamp_s_to_ms(source_column, name):
    """
    This Constructor is used for unix timestamps. The values are cleaned
    next to casting and renaming.
    If the values are not between `01.01.1970` and `31.12.2099`,
    NULL will be returned.
    Cast to :any:`pyspark.sql.types.LongType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(time_sec=1581540839),  # 02/12/2020 @ 8:53pm (UTC)
    >>>     Row(time_sec=-4887839),    # Invalid!
    >>>     Row(time_sec=4737139200)   # 02/12/2120 @ 12:00am (UTC)
    >>> ])
    >>>
    >>> mapping = [("unix_ts", "time_sec", "timestamp_s_to_ms")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(unix_ts=1581540839000), Row(unix_ts=None), Row(unix_ts=None)]

    Note
    ----
    *input*  in **seconds**
    *output* in **milli seconds**
    """

    return (
        F.when(
            source_column.between(MIN_TIMESTAMP_SEC, MAX_TIMESTAMP_SEC),
            source_column * 1000,
        )
        .otherwise(F.lit(None))
        .cast(T.LongType())
        .alias(name)
    )


def _generate_select_expression_for_timestamp_s_to_s(source_column, name):
    """
    This Constructor is used for unix timestamps. The values are cleaned
    next to casting and renaming.
    If the values are not between `01.01.1970` and `31.12.2099`,
    NULL will be returned.
    Cast to :any:`pyspark.sql.types.LongType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(time_sec=1581540839),  # 02/12/2020 @ 8:53pm (UTC)
    >>>     Row(time_sec=-4887839),    # Invalid!
    >>>     Row(time_sec=4737139200)   # 02/12/2120 @ 12:00am (UTC)
    >>> ])
    >>>
    >>> mapping = [("unix_ts", "time_sec", "timestamp_s_to_ms")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(unix_ts=1581540839), Row(unix_ts=None), Row(unix_ts=None)]

    Note
    ----
    *input*  in **seconds**
    *output* in **seconds**
    """
    return (
        F.when(source_column.between(MIN_TIMESTAMP_SEC, MAX_TIMESTAMP_SEC), source_column)
        .otherwise(F.lit(None))
        .cast(T.LongType())
        .alias(name)
    )


def _generate_select_expression_for_StringNull(source_column, name):  # noqa: N802
    """
    Used for Anonymizing.
    Input values will be ignored and replaced by NULL,
    Cast to :any:`pyspark.sql.types.StringType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(email=u'tsivorn1@who.int'),
    >>>      Row(email=u''),
    >>>      Row(email=u'gisaksen4@skype.com')]
    >>> )
    >>>
    >>> mapping = [("email", "email", "StringNull")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(email=None), Row(email=None), Row(email=None)]
    """

    return F.lit(None).cast(T.StringType()).alias(name)


def _generate_select_expression_for_IntNull(source_column, name):  # noqa: N802
    """
    Used for Anonymizing.
    Input values will be ignored and replaced by NULL,
    Cast to :any:`pyspark.sql.types.IntegerType`

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(facebook_id=3047288),
    >>>      Row(facebook_id=0),
    >>>      Row(facebook_id=57815)]
    >>> )
    >>>
    >>> mapping = [("facebook_id", "facebook_id", "IntNull")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(facebook_id=None), Row(facebook_id=None), Row(facebook_id=None)]
    """

    return F.lit(None).cast(T.IntegerType()).alias(name)


def _generate_select_expression_for_StringBoolean(source_column, name):  # noqa: N802
    """
    Used for Anonymizing.
    The column's value will be replaced by `"1"` if it is:

        * not NULL and
        * not an empty string

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(email=u'tsivorn1@who.int'),
    >>>      Row(email=u''),
    >>>      Row(email=u'gisaksen4@skype.com')]
    >>> )
    >>>
    >>> mapping = [("email", "email", "StringBoolean")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(email=u'1'), Row(email=None), Row(email=u'1')]
    """
    return (
        F.when(source_column.isNull(), F.lit(None))
        .when(source_column == "", F.lit(None))
        .otherwise("1")
        .cast(T.StringType())
        .alias(name)
    )


def _generate_select_expression_for_IntBoolean(source_column, name):  # noqa: N802
    """
    Used for Anonymizing.
    The column's value will be replaced by `1` if it contains a non-NULL value.

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(facebook_id=3047288),
    >>>      Row(facebook_id=0),
    >>>      Row(facebook_id=None)]
    >>> )
    >>>
    >>> mapping = [("facebook_id", "facebook_id", "IntBoolean")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(facebook_id=1), Row(facebook_id=1), Row(facebook_id=None)]

    Note
    ----
    `0` (zero) or negative numbers are still considered as valid values and therefore converted to `1`.
    """
    return F.when(source_column.isNull(), F.lit(None)).otherwise(1).cast(T.IntegerType()).alias(name)


def _generate_select_expression_for_TimestampMonth(source_column, name):  # noqa: N802
    """
    Used for Anonymizing. Can be used to keep the age but obscure the explicit birthday.
    This custom datatype requires a :any:`pyspark.sql.types.TimestampType` column as input.
    The datetime value will be set to the first day of the month.

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from datetime import datetime
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(birthday=datetime(2019, 2, 9, 2, 45)),
    >>>      Row(birthday=None),
    >>>      Row(birthday=datetime(1988, 1, 31, 8))]
    >>> )
    >>>
    >>> mapping = [("birthday", "birthday", "TimestampMonth")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(birthday=datetime.datetime(2019, 2, 1, 0, 0)),
     Row(birthday=None),
     Row(birthday=datetime.datetime(1988, 1, 1, 0, 0))]
    """
    return F.trunc(source_column, "month").cast(T.TimestampType()).alias(name)


def _generate_select_expression_for_meters_to_cm(source_column, name):
    """
    Convert meters to cm and cast the result to an IntegerType.

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(size_in_m=1.80),
    >>>      Row(size_in_m=1.65),
    >>>      Row(size_in_m=2.05)]
    >>> )
    >>>
    >>> mapping = [("size_in_cm", "size_in_m", "meters_to_cm")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(size_in_cm=180),
     Row(size_in_cm=165),
     Row(size_in_cm=205)]
    """
    return (source_column * 100).cast(T.IntegerType()).alias(name)


def _generate_select_expression_for_has_value(source_column, name):
    """
    Returns True if the source_column is
        - not NULL and
        - not "" (empty string)

    otherwise it returns False

    Warning
    -------
    This means that it will return True for values which would indicate a False value. Like "false" or 0!!!

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(input_key=1.80),
    >>>      Row(input_key=None),
    >>>      Row(input_key="some text"),
    >>>      Row(input_key="")]
    >>> )
    >>>
    >>> mapping = [("input_key", "result", "has_value")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(4)
    [Row(result=True),
     Row(result=False),
     Row(result=True),
     Row(result=False)]

    """
    return (
        F.when((source_column.isNotNull()) & (source_column.cast(T.StringType()) != ""), F.lit(True))
        .otherwise(F.lit(False))
        .alias(name)
    )


def _generate_select_expression_for_unix_timestamp_ms_to_spark_timestamp(source_column, name):
    """
    Convert unix timestamps in milliseconds to a Spark TimeStampType. It is assumed that the
    timezone is already set to UTC in spark / java to avoid implicit timezone conversions.

    Example
    -------
    >>> from pyspark.sql import Row
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df = spark.createDataFrame(
    >>>     [Row(unix_timestamp_in_ms=1591627696951),
    >>>      Row(unix_timestamp_in_ms=1596812952000),
    >>>      Row(unix_timestamp_in_ms=946672200000)]
    >>> )
    >>>
    >>> mapping = [("spark_timestamp", "unix_timestamp_in_ms", "unix_timestamp_ms_to_spark_timestamp")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(spark_timestamp=datetime.datetime(2020, 6, 8, 16, 48, 16, 951000)),
     Row(spark_timestamp=datetime.datetime(2020, 8, 7, 17, 9, 12)),
     Row(spark_timestamp=datetime.datetime(1999, 12, 31, 21, 30))]
    """
    return (source_column / 1000).cast(T.TimestampType()).alias(name)


def _generate_select_expression_for_extended_string_to_int(source_column, name):
    """
    More robust conversion from StringType to IntegerType.
    Is able to additionally handle (compared to implicit Spark conversion):

        * Preceding whitespace
        * Trailing whitespace
        * Preceeding and trailing whitespace
        * underscores as thousand separators

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="  123456 "),
     Row(input_string="Hello"),
     Row(input_string="123_456")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_int")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=123456),
     Row(input_string=None),
     Row(input_string=123456)]
    """
    return _generate_select_expression_for_extended_string_to_long(source_column, name).cast(T.IntegerType())


def _generate_select_expression_for_extended_string_to_long(source_column, name):
    """
    More robust conversion from StringType to LongType.
    Is able to additionally handle (compared to implicit Spark conversion):

    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace
    * underscores as thousand separators

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="  21474836470 "),
     Row(input_string="Hello"),
     Row(input_string="21_474_836_470")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_long")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=21474836470),
     Row(input_string=None),
     Row(input_string=21474836470)]
    """
    return F.regexp_replace(F.trim(source_column), "_", "").cast(T.LongType()).alias(name)


def _generate_select_expression_for_extended_string_to_float(source_column, name):
    """
    More robust conversion from StringType to FloatType.
    Is able to additionally handle (compared to implicit Spark conversion):

    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace
    * underscores as thousand separators

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="  836470.819 "),
     Row(input_string="Hello"),
     Row(input_string="836_470.819")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_float")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=836470.819),
     Row(input_string=None),
     Row(input_string=836470.819)]
    """
    return _generate_select_expression_for_extended_string_to_double(source_column, name).cast(T.FloatType())


def _generate_select_expression_for_extended_string_to_double(source_column, name):
    """
    More robust conversion from StringType to DoubleType.
    Is able to additionally handle (compared to implicit Spark conversion):

    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace
    * underscores as thousand separators

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="  21474838464.70 "),
     Row(input_string="Hello"),
     Row(input_string="21_474_838_464.70")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_double")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=21474838464.7),
     Row(input_string=None),
     Row(input_string=21474838464.70)]
    """
    return F.regexp_replace(F.trim(source_column), "_", "").cast(T.DoubleType()).alias(name)


def _generate_select_expression_for_extended_string_to_boolean(source_column, name):
    """
    More robust conversion from StringType to BooleanType.
    Is able to additionally handle (compared to implicit Spark conversion):

    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace

    Warning
    ---------
    This does not handle numbers (cast as string) the same way as numbers (cast as number) to boolean conversion!
    F.e.

    * 100 to boolean => True
    * "100" to extended_string_to_boolean => False
    * "100" to boolean => False

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="  true "),
     Row(input_string="0"),
     Row(input_string="y")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_boolean")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=True),
     Row(input_string=False),
     Row(input_string=True)]
    """
    true_values = ["on", "enabled"]
    false_values = ["off", "disabled"]
    return (
        F.when(F.trim(source_column).isin(true_values), F.lit(True))
        .when(F.trim(source_column).isin(false_values), F.lit(False))
        .otherwise(F.trim(source_column).cast(T.BooleanType()))
    ).alias(name)


def _generate_select_expression_for_extended_string_to_timestamp(source_column, name):
    """
    More robust conversion from StringType to TimestampType. It is assumed that the
    timezone is already set to UTC in spark / java to avoid implicit timezone conversions.

    The conversion can handle unix timestamps in seconds and in milliseconds:
        - Timestamps in the range [-MAX_TIMESTAMP_S, MAX_TIMESTAMP_S] are treated as seconds
        - Timestamps in the range [-inf, -MAX_TIMESTAMP_S) and (MAX_TIMESTAMP_S, inf] are treated as milliseconds
        - There is a time interval (1970-01-01 +- ~2.5 months)where we can not distinguish correctly between s and ms
          (e.g. 3974400000 would be treated as seconds (2095-12-11T00:00:00) as the value is smaller than
          MAX_TIMESTAMP_S, but it could also be a valid date in Milliseconds (1970-02-16T00:00:00)

    Is able to additionally handle (compared to implicit Spark conversion):
    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="2020-08-12T12:43:14+0000"),
     Row(input_string="1597069446"),
     Row(input_string="2020-08-12")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_timestamp")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=datetime.datetime(2020, 8, 12, 12, 43, 14)),
     Row(input_string=datetime.datetime(2020, 8, 10, 14, 24, 6)),
     Row(input_string=datetime.datetime(2020, 8, 12, 0, 0, 0))]
    """
    return (
        F.when(
            F.abs(F.trim(source_column).cast(T.LongType())).between(0, MAX_TIMESTAMP_SEC),
            F.trim(source_column).cast(T.LongType()).cast(T.TimestampType()),
        )
        .when(
            F.abs(F.trim(source_column).cast(T.LongType())) > MAX_TIMESTAMP_SEC,
            (F.trim(source_column) / 1000).cast(T.TimestampType()),
        )
        .otherwise(F.trim(source_column).cast(T.TimestampType()))
        .alias(name)
    )


def _generate_select_expression_for_extended_string_to_date(source_column, name):
    """
    More robust conversion from StringType to DateType. It is assumed that the
    timezone is already set to UTC in spark / java to avoid implicit timezone conversions.

    The conversion can handle unix timestamps in seconds and in milliseconds:
        - Timestamps in the range [-MAX_TIMESTAMP_S, MAX_TIMESTAMP_S] are treated as seconds
        - Timestamps in the range [-inf, -MAX_TIMESTAMP_S) and (MAX_TIMESTAMP_S, inf] are treated as milliseconds
        - There is a time interval (1970-01-01 +- ~2.5 months)where we can not distinguish correctly between s and ms
          (e.g. 3974400000 would be treated as seconds (2095-12-11T00:00:00) as the value is smaller than
          MAX_TIMESTAMP_S, but it could also be a valid date in Milliseconds (1970-02-16T00:00:00)

    Is able to additionally handle (compared to implicit Spark conversion):
    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="2020-08-12T12:43:14+0000"),
     Row(input_string="1597069446"),
     Row(input_string="2020-08-12")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_date")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=datetime.datetime(2020, 8, 12)),
     Row(input_string=datetime.datetime(2020, 8, 10)),
     Row(input_string=datetime.datetime(2020, 8, 12))]
    """
    return _generate_select_expression_for_extended_string_to_timestamp(source_column, name).cast(T.DateType())


def _generate_select_expression_for_extended_string_unix_timestamp_ms_to_timestamp(source_column, name):
    """
    More robust conversion from StringType to TimestampType. It is assumed that the
    timezone is already set to UTC in spark / java to avoid implicit timezone conversions.
    Is able to additionally handle (compared to implicit Spark conversion):

    * Unix timestamps in milliseconds
    * Preceding whitespace
    * Trailing whitespace
    * Preceeding and trailing whitespace

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="2020-08-12T12:43:14+0000"),
     Row(input_string="1597069446000"),
     Row(input_string="2020-08-12")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_timestamp")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=datetime.datetime(2020, 8, 12, 12, 43, 14)),
     Row(input_string=datetime.datetime(2020, 8, 10, 14, 24, 6)),
     Row(input_string=datetime.datetime(2020, 8, 12, 0, 0, 0))]
    """
    return (
        F.when(
            F.trim(source_column).cast(T.LongType()).isNotNull(), (F.trim(source_column) / 1000).cast(T.TimestampType())
        )
        .otherwise(F.trim(source_column).cast(T.TimestampType()))
        .alias(name)
    )


def _generate_select_expression_for_extended_string_unix_timestamp_ms_to_date(source_column, name):
    """
    More robust conversion from StringType to DateType. It is assumed that the
    timezone is already set to UTC in spark / java to avoid implicit timezone conversions and that
    unix timestamps are in **milli seconds**

    Hint
    ----
    Please have a look at the tests to get a better feeling how it behaves under
    tests/unit/transformer/test_mapper_custom_data_types.py::TestExtendedStringConversions and
    tests/data/test_fixtures/mapper_custom_data_types_fixtures.py

    Example
    -------
    >>> from spooq.transformer import Mapper
    >>>
    >>> input_df.head(3)
    [Row(input_string="2020-08-12T12:43:14+0000"),
     Row(input_string="1597069446000"),
     Row(input_string="2020-08-12")]
    >>> mapping = [("output_value", "input_string", "extended_string_to_date")]
    >>> output_df = Mapper(mapping).transform(input_df)
    >>> output_df.head(3)
    [Row(input_string=datetime.datetime(2020, 8, 12)),
     Row(input_string=datetime.datetime(2020, 8, 10)),
     Row(input_string=datetime.datetime(2020, 8, 12))]
    """
    return _generate_select_expression_for_extended_string_unix_timestamp_ms_to_timestamp(source_column, name).cast(
        T.DateType()
    )



from builtins import str
import os
import logging

spooq_logger = logging.getLogger("spooq")


def remove_hdfs_prefix(path):
    """Replaces references to HDFS from the input path

    Example
    -------
    >>> _remove_hdfs_prefix(u'hdfs://nameservice-ha:8020/user/furia_salamandra_faerfax/data')
    u'/user/furia_salamandra_faerfax/data'
    """
    cleaned_path = path.replace("hdfs://nameservice-ha", "").replace(":8020", "")
    spooq_logger.debug("remove_hdfs_prefix: input: {inp}, output: {outp}".format(inp=path, outp=cleaned_path))
    return cleaned_path


def fix_suffix(path):
    """
    Ensures Full Input Path ends with ``/*``

    Example
    -------
    >>> _fix_suffix(u'/user/furia_salamandra_faerfax/data/18/01/01')
    u'/user/furia_salamandra_faerfax/data/18/01/01/*'

    """
    if path[-2:] == "/*":
        cleaned_path = path
    else:
        cleaned_path = os.path.join(path, "*")
    spooq_logger.debug("fix_suffix: input: {inp}, output: {outp}".format(inp=path, outp=cleaned_path))
    return cleaned_path


def infer_input_path_from_partition(base_path=None, partition=None):
    """
    Derives the final input_path from the cleaned :py:data:`base_path` & :py:data:`partition`
    and returns it.

    Parameters
    ----------
    base_path : :any:`str`
    partition : :any:`str` or :any:`int`

    Returns
    -------
    :any:`str`
        The final input_path to be used for Extraction.

    See Also
    --------
    _clean_path
    """
    base_path = remove_hdfs_prefix(base_path)
    partition = str(partition)
    inferred_path = "{base_path}/{yy}/{mm}/{dd}/*".format(
        base_path=base_path, yy=partition[2:4], mm=partition[4:6], dd=partition[6:8]
    )
    spooq_logger.debug(
        "infer_input_path_from_partition: input: {inp}, output: {outp}".format(inp=base_path, outp=inferred_path)
    )
    return inferred_path



from .newest_by_group import NewestByGroup
from .mapper import Mapper
from .exploder import Exploder
from .threshold_cleaner import ThresholdCleaner
from .enum_cleaner import EnumCleaner
from .sieve import Sieve

__all__ = [
    "NewestByGroup",
    "Mapper",
    "Exploder",
    "ThresholdCleaner",
    "EnumCleaner",
    "Sieve",
]



from __future__ import absolute_import
from builtins import str
from pyspark.sql.utils import AnalysisException
from pyspark.sql import functions as F
from pyspark.sql import types as T
from pyspark.sql.column import Column

from .transformer import Transformer
from .mapper_custom_data_types import _get_select_expression_for_custom_type


class Mapper(Transformer):
    """
    Constructs and applies a PySpark SQL expression, based on the provided mapping.

    Examples
    --------
    >>> from pyspark.sql import functions as F
    >>> from spooq.transformer import Mapper
    >>>
    >>> mapping = [
    >>>     ("id",            "data.relationships.food.data.id",  "StringType"),
    >>>     ("version",       "data.version",                     "extended_string_to_int"),
    >>>     ("type",          "elem.attributes.type",             "StringType"),
    >>>     ("created_at",    "elem.attributes.created_at",       "extended_string_to_timestamp"),
    >>>     ("created_on",    "elem.attributes.created_at",       "extended_string_to_date"),
    >>>     ("process_date",  F.current_timestamp(),              "DateType"),
    >>> ]
    >>> mapper = Mapper(mapping=mapping)
    >>> mapper.transform(input_df).printSchema()
    root
     |-- id: string (nullable = true)
     |-- version: integer (nullable = true)
     |-- type: string (nullable = true)
     |-- created_at: timestamp (nullable = true)
     |-- created_on: date (nullable = true)
     |-- process_date: date (nullable = false)


    Parameters
    ----------
    mapping  : :class:`list` of :any:`tuple` containing three :any:`str` or :class:`~pyspark.sql.Column` or :mod:`~pyspark.sql.functions`
        This is the main parameter for this transformation. It gives information
        about the column names for the output DataFrame, the column names (paths)
        from the input DataFrame, and their data types. Custom data types are also supported, which can
        clean, pivot, anonymize, ... the data itself. Please have a look at the
        :py:mod:`spooq.transformer.mapper_custom_data_types` module for more information.

    ignore_missing_columns : :any:`bool`, Defaults to False
        Specifies if the mapping transformation should use NULL if a referenced input
        column is missing in the provided DataFrame. If set to False, it will raise an exception.

    ignore_ambiguous_columns : :any:`bool`, Defaults to False
        It can happen that the input DataFrame has ambiguous column names (like "Key" vs "key") which will
        raise an exception with Spark when reading. This flag surpresses this exception and skips those affected
        columns.

    mode : :any:`str`, Defaults to "replace"
        Defines weather the mapping should fully replace the schema of the input DataFrame or just add to it.
        Following modes are supported:

            * replace
                The output schema is the same as the provided mapping.
                => output schema: new columns
            * append
                The columns provided in the mapping are added at the end of the input schema. If a column already
                exists in the input DataFrame, its position is kept.
                => output schema: input columns + new columns
            * prepend
                The columns provided in the mapping are added at the beginning of the input schema. If a column already
                exists in the input DataFrame, its position is kept.
                => output schema: new columns + input columns

    Note
    ----
    Let's talk about Mappings:

    The mapping should be a list of tuples that contain all necessary information per column.

    * Column Name: :any:`str`
        Sets the name of the column in the resulting output DataFrame.
    * Source Path / Name / Column / Function: :any:`str` or :class:`~pyspark.sql.Column` or :mod:`~pyspark.sql.functions`
        Points to the name of the column in the input DataFrame. If the input
        is a flat DataFrame, it will essentially be the column name. If it is of complex
        type, it will point to the path of the actual value. For example: ``data.relationships.sample.data.id``,
        where id is the value we want. It is also possible to directly pass
        a PySpark Column which will get evaluated. This can contain arbitrary logic supported by Spark. For example:
        ``F.current_date()`` or ``F.when(F.col("size") == 180, F.lit("tall")).otherwise(F.lit("tiny"))``.
    * DataType: :any:`str` or :class:`~pyspark.sql.types.DataType`
        DataTypes can be types from :any:`pyspark.sql.types`, selected custom datatypes or
        injected, ad-hoc custom datatypes.
        The datatype will be interpreted as a PySpark built-in if it is a member of :any:`pyspark.sql.types` module.
        If it is not an importable PySpark data type, a method to construct the statement will be
        called by the data type's name.

    Note
    ----
    The available input columns can vary from batch to batch if you use schema inference
    (f.e. on json data) for the extraction. Ignoring missing columns on the input DataFrame is
    highly encouraged in this case. Although, if you have tight control over the structure
    of the extracted DataFrame, setting `ignore_missing_columns` to True is advised
    as it can uncover typos and bugs.

    Note
    ----
    Please see :py:mod:`spooq.transformer.mapper_custom_data_types` for all available custom
    data types and how to inject your own.

    Note
    ----
    Attention: Decimal is NOT SUPPORTED by Hive! Please use Double instead!
    """

    def __init__(self, mapping, ignore_missing_columns=False, ignore_ambiguous_columns=False, mode="replace"):
        super(Mapper, self).__init__()
        self.mapping = mapping
        self.ignore_missing_columns = ignore_missing_columns
        self.ignore_ambiguous_columns = ignore_ambiguous_columns
        self.mode = mode

    def transform(self, input_df):
        self.logger.info("Generating SQL Select-Expression for Mapping...")
        self.logger.debug("Input Schema/Mapping:")
        self.logger.debug("\n" + "\n".join(["\t".join(map(str, mapping_line)) for mapping_line in self.mapping]))

        input_columns = input_df.columns
        select_expressions = []
        with_column_expressions = []

        for (name, source_column, data_type) in self.mapping:
            self.logger.debug("generating Select statement for attribute: {nm}".format(nm=name))

            source_column = self._get_spark_column(source_column, name, input_df)
            if source_column is None:
                continue
            data_type, data_type_is_spark_builtin = self._get_spark_data_type(data_type)
            select_expression = self._get_select_expression(name, source_column, data_type, data_type_is_spark_builtin)

            self.logger.debug(
                "Select-Expression for Attribute {nm}: {sql_expr}".format(nm=name, sql_expr=str(select_expression))
            )
            if self.mode != "replace" and name in input_columns:
                with_column_expressions.append((name, select_expression))
            else:
                select_expressions.append(select_expression)

        self.logger.info("SQL Select-Expression for new mapping generated!")
        self.logger.debug("SQL Select-Expressions for new mapping:\n" + "\n".join(str(select_expressions).split(",")))
        self.logger.debug("SQL WithColumn-Expressions for new mapping: " + str(with_column_expressions))
        if self.mode == "prepend":
            df_to_return = input_df.select(select_expressions + ["*"])
        elif self.mode == "append":
            df_to_return = input_df.select(["*"] + select_expressions)
        elif self.mode == "replace":
            df_to_return = input_df.select(select_expressions)
        else:
            exception_message = (
                "Only 'prepend', 'append' and 'replace' are allowed for Mapper mode!"
                "Value: '{val}' was used as mode for the Mapper transformer."
            )
            self.logger.exception(exception_message)
            raise ValueError(exception_message)

        if with_column_expressions:
            for name, expression in with_column_expressions:
                df_to_return = df_to_return.withColumn(name, expression)
        return df_to_return

    def _get_spark_column(self, source_column, name, input_df):
        """
        Returns the provided source column as a Pyspark.sql.Column and marks if it is missing or not.
        Supports source column definition as a string or a Pyspark.sql.Column (including functions).
        """
        try:
            input_df.select(source_column)
            if isinstance(source_column, str):
                source_column = F.col(source_column)

        except AnalysisException as e:
            if isinstance(source_column, str) and self.ignore_missing_columns:
                self.logger.warn(
                    f"Missing column ({str(source_column)}) replaced with NULL (via ignore_missing_columns=True): {e.desc}"
                )
                source_column = F.lit(None)
            elif "ambiguous" in e.desc.lower() and self.ignore_ambiguous_columns:
                self.logger.warn(
                    f'Exception ignored (via ignore_ambiguous_columns=True) for column "{str(source_column)}": {e.desc}'
                )
                return None
            else:
                self.logger.exception(
                    'Column: "{}" cannot be resolved '.format(str(source_column))
                    + 'but is referenced in the mapping by column: "{}".\n'.format(name)
                )
                raise e
        return source_column

    @staticmethod
    def _get_spark_data_type(data_type):
        """
        Returns the provided data_type as a Pyspark.sql.type.DataType (for spark-built-ins)
        or as a string (for custom spooq transformations) and marks if it is built-in or not.
        Supports source column definition as a string or a Pyspark.sql.Column (including functions).
        """
        if isinstance(data_type, T.DataType):
            data_type_is_spark_builtin = True
        elif isinstance(data_type, str):
            data_type = data_type.replace("()", "")
            if hasattr(T, data_type):
                data_type_is_spark_builtin = True
                data_type = getattr(T, data_type)()
            else:
                data_type_is_spark_builtin = False
        else:
            raise ValueError(
                'data_type not supported! class: "{}", name: "{}"'.format(type(data_type).__name__, str(data_type))
            )
        return data_type, data_type_is_spark_builtin

    @staticmethod
    def _get_select_expression(name, source_column, data_type, data_type_is_spark_builtin):
        """
        Returns a valid pyspark sql select-expression with cast and alias, depending on the input parameters.
        """
        if data_type_is_spark_builtin:
            return source_column.cast(data_type).alias(name)
        else:  # Custom Data Type
            return _get_select_expression_for_custom_type(source_column, name, data_type)



"""
Transformers take a :py:class:`pyspark.sql.DataFrame` as an input, transform it accordingly
and return a :py:class:`pyspark.sql.DataFrame`.

Each Transformer class has to have a `transform` method which takes no arguments
and returns a :py:class:`pyspark.sql.DataFrame`.

Possible transformation methods can be Selecting the most up to date record by id,
Exploding an array, Filter (on an exploded array), Apply basic threshold cleansing or
Map the incoming DataFrame to at provided structure.
"""

from builtins import object
import logging


class Transformer(object):
    """
    Base Class of Transformer Classes.

    Attributes
    ----------
    name : :any:`str`
        Sets the `__name__` of the class' type as `name`, which is essentially the Class' Name.
    logger : :any:`logging.Logger`
        Shared, class level logger for all instances.
    """

    def __init__(self):
        self.name = type(self).__name__
        self.logger = logging.getLogger("spooq")

    def transform(self, input_df):
        """
        Performs a transformation on a DataFrame.

        Parameters
        ----------
        input_df : :py:class:`pyspark.sql.DataFrame`
            Input DataFrame

        Returns
        -------
        :py:class:`pyspark.sql.DataFrame`
            Transformed DataFrame.

        Note
        ----
        This method does only take the Input DataFrame as a parameters. All other needed parameters
        are defined in the initialization of the Transformator Object.
        """
        raise NotImplementedError("This method has to be implemented in the subclasses")

    def __str__(self):
        return "Transformer Object of Class {nm}".format(nm=self.name)



from __future__ import absolute_import
import sys

if sys.version_info.major > 2:
    # This is needed for python 2 as otherwise pyspark raises an exception for following command:
    # data_type = input_df.schema[str(column_name)].dataType
    # Pyspark checks if the input is a string, which does not work
    # with the new strings from builtins
    from builtins import str
import pyspark.sql.functions as F
import pyspark.sql.types as sql_types
from pyspark.sql.column import Column

from .base_cleaner import BaseCleaner


class ThresholdCleaner(BaseCleaner):
    """
    Sets outiers within a DataFrame to a default value.
    Takes a dictionary with valid value ranges for each column to be cleaned.

    Examples
    --------
    >>> from pyspark.sql import functions as F
    >>> from spooq.transformer import ThresholdCleaner
    >>> transformer = ThresholdCleaner(
    >>>     thresholds={
    >>>         "created_at": {
    >>>             "min": 0,
    >>>             "max": 1580737513,
    >>>             "default": F.current_date()
    >>>         },
    >>>         "size_cm": {
    >>>             "min": 70,
    >>>             "max": 250,
    >>>             "default": None
    >>>         },
    >>>     }
    >>> )

    >>> from spooq.transformer import ThresholdCleaner
    >>> from pyspark.sql import Row
    >>>
    >>> input_df = spark.createDataFrame([
    >>>     Row(id=0, integers=-5, some_factor=-0.75),
    >>>     Row(id=1, integers= 5, some_factor= 1.25),
    >>>     Row(id=2, integers=15, some_factor= 0.67),
    >>> ])
    >>> transformer = ThresholdCleaner(
    >>>     thresholds={
    >>>         "integers":    {"min":  0, "max": 10},
    >>>         "some_factor": {"min": -1, "max":  1}
    >>>     },
    >>>     column_to_log_cleansed_values="cleansed_values_threshold",
    >>>     store_as_map=True,
    >>> )
    >>> output_df = transformer.transform(input_df)
    >>> output_df.show()
    +---+--------+-----------+-------------------------+
    | id|integers|some_factor|cleansed_values_threshold|
    +---+--------+-----------+-------------------------+
    |  0|    null|      -0.75|         [integers -> -5]|
    |  1|       5|       null|     [some_factor -> 1...|
    |  2|    null|       0.67|         [integers -> 15]|
    +---+--------+-----------+-------------------------+
    #  Remark: ``some_factor -> 1...`` from the df.show() is a bug, the stored value correctly holds ``"1.25"``
    >>> output_df.printSchema()
     |-- id: long (nullable = true)
     |-- integers: long (nullable = true)
     |-- some_factor: double (nullable = true)
     |-- cleansed_values_threshold: map (nullable = false)
     |    |-- key: string
     |    |-- value: string (valueContainsNull = true)

    Parameters
    ----------
    thresholds : :py:class:`dict`
        Dictionary containing column names and respective valid ranges

    column_to_log_cleansed_values : :any:`str`, Defaults to None
        Defines a column in which the original (uncleansed) value will be stored in case of cleansing. If no column
        name is given, nothing will be logged.

    store_as_map : :any:`bool`, Defaults to False
        Specifies if the logged cleansed values should be stored in a column as :any:`pyspark.sql.types.MapType` with
        stringified values or as :any:`pyspark.sql.types.StructType` with the original respective data types.

    Note
    ----
    Following cleansing rule attributes per column are supported:

        * min, mandatory - any
            A number or timestamp/date which serves as the lower limit for allowed values. Values
            below this threshold will be cleansed.
        * max, mandatory - any
            A number or timestamp/date which serves as the upper limit for allowed values. Values
            above this threshold will be cleansed.
        * default, defaults to None - :class:`~pyspark.sql.column.Column` or any primitive Python value
            If a value gets cleansed it gets replaced with the provided default value.

    The :meth:`pyspark.sql.functions.between` method is used internally.

    Returns
    -------
    :any:`pyspark.sql.DataFrame`
        The transformed DataFrame

    Raises
    ------
    :any:`exceptions.ValueError`
        Threshold-based cleaning only supports Numeric, Date and Timestamp Types!
        Column with name: {col_name} and type of: {col_type} was provided

    Warning
    -------
    Only Numeric, TimestampType, and DateType data types are supported!
    """

    def __init__(self, thresholds={}, column_to_log_cleansed_values=None, store_as_map=False):
        super().__init__(
            cleaning_definitions=thresholds,
            column_to_log_cleansed_values=column_to_log_cleansed_values,
            store_as_map=store_as_map,
            temporary_columns_prefix="dac28b56d8055953a7038bfe3b5097e7",
        )
        self.logger.debug("Range Definitions: " + str(self.cleaning_definitions))

    def transform(self, input_df):
        self.logger.debug("input_df Schema: " + input_df._jdf.schema().treeString())
        ordered_column_names = input_df.columns

        if self.column_to_log_cleansed_values:
            column_names_to_clean = self.cleaning_definitions.keys()
            temporary_column_names = self._get_temporary_column_names(column_names_to_clean)
            input_df = self._add_temporary_columns(input_df, column_names_to_clean, temporary_column_names)

        cleansing_expressions = []
        for column_name, value_range in list(self.cleaning_definitions.items()):

            data_type = input_df.schema[str(column_name)].dataType
            substitute = value_range.get("default", None)
            if not isinstance(substitute, Column):
                substitute = F.lit(substitute)

            if not isinstance(data_type, (sql_types.NumericType, sql_types.DateType, sql_types.TimestampType)):
                raise ValueError(
                    "Threshold-based cleaning only supports Numeric, Date and Timestamp Types!\n",
                    "Column with name: {col_name} and type of: {col_type} was provided".format(
                        col_name=column_name, col_type=data_type
                    ),
                )

            self.logger.debug("Ranges for column " + column_name + ": " + str(value_range))

            cleansing_expression = (
                F.when(
                    input_df[column_name].between(value_range["min"], value_range["max"]),
                    input_df[column_name],
                )
                .otherwise(substitute)
                .cast(data_type)
            )
            self.logger.debug("Cleansing Expression for " + column_name + ": " + str(cleansing_expression))
            cleansing_expressions.append((column_name, cleansing_expression))

        self.logger.info("Full threshold cleansing expression:")
        self.logger.info(
            ".".join(
                [
                    f"withColumn({column_name}, {str(cleansing_expr)})"
                    for (column_name, cleansing_expr) in cleansing_expressions
                ]
            )
        )
        for (column_name, cleansing_expr) in cleansing_expressions:
            input_df = input_df.withColumn(column_name, cleansing_expr)

        if self.column_to_log_cleansed_values:
            input_df = self._log_cleansed_values(input_df, column_names_to_clean, temporary_column_names)
            ordered_column_names.append(self.column_to_log_cleansed_values)

        return input_df.select(ordered_column_names)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import re
from setuptools import setup, find_packages

version_filename = "spooq/_version.py"
with open(version_filename, "rt") as version_filename:
    version_file = version_filename.read()
    version_regex    = r"""^__version__ = ["\"]([^"\"]*)["\"]"""
    regex_result     = re.search(version_regex, version_file, re.M)
    if regex_result:
        version_string = regex_result.group(1)
    else:
        raise RuntimeError("Unable to find version string in %s." % (version_filename,))

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("CHANGELOG.rst") as history_file:
    history = history_file.read()

with open("docs/source/index.rst") as sphinx_index_file:
    sphinx_index = sphinx_index_file.read()

requirements = [
    "pandas",
    "future",
    "requests",
]

setup(
    name="Spooq",
    version=version_string,
    description="""
        Spooq is a PySpark based helper library for ETL data ingestion pipeline in Data Lakes.
    """,
    long_description=readme + "\n\n" + history,
    # long_description=sphinx_index,
    long_description_content_type="text/markdown",
    author="David Hohensinn",
    author_email="breaka@gmx.at",
    url="https://github.com/Breaka84/Spooq",
    project_urls={
        "Documentation": "https://spooq.readthedocs.io",

    },
    include_package_data=False,
    packages=find_packages(exclude=['*tests*']),
    install_requires=requirements,
    zip_safe=False,
    keywords=[
        "spooq", "spark", "hive", "cloudera", "hadoop", "etl",
        "data ingestion", "data wrangling", "databricks", "big data",
        "batch", "streaming"
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ]
)



from .jdbc import JDBCExtractorIncremental, JDBCExtractorFullLoad
from .json_files import JSONExtractor

__all__ = [
    "JDBCExtractorIncremental",
    "JDBCExtractorFullLoad",
    "JSONExtractor",
]
